#include "extension.h"
#include <ISmmPlugin.h>
ISmmPlugin *mmplugin;
// sigscanner stuff
#include "sigscan_base.h"
#include "itempents.h"
// recipient filter
#include "MRecipientFilter.h"
#include "eiface.h"

TempentsExtension g_TempentsExtension;
SMEXT_LINK(&g_TempentsExtension);

// define our signature stuff here
#ifdef WIN32
	SigScan TempentsSig;
#endif

// we need an interface for te
ITempEntsSystem *te=NULL;

bool TempentsExtension::SDK_OnLoad(char *error, size_t err_max, bool late)
{
	g_pShareSys->AddNatives(myself, tempents_natives);
	return true;
}

void TempentsExtension::SDK_OnUnload()
{
}

bool TempentsExtension::SDK_OnMetamodLoad(ISmmAPI *ismm, char *error, size_t maxlen, bool late)
{
	mmplugin=this;
#ifdef WIN32
	Ext_SigScan::Init_Sigs();
	TempentsSig.Init("DispatchEffect",
	(unsigned char*)"\x83\xEC\x20\x56\x8D\x4C\x24\x04\xE8\xD3\xBF\xFB\xFF\x8B\x74\x24",
	"xxxxxxxxx?xxxxxx",
	16);
	if(TempentsSig.is_set)
	{
		memcpy(&te, ((char*)TempentsSig.sig_addr + 0x29), sizeof(char*));
		int *ptr = *((int **)(te));
		te = (ITempEntsSystem *)ptr;
	}
	else
		g_SMAPI->ConPrintf("** ERROR ** Tempents extension could not load correctly!\n");
#else
	void *pHandle, *pAddr;
	char binpath[512];
	engine->GetGameDir(binpath,512);
	sprintf(binpath, "%s/bin/server_i486.so",binpath);
	if((pHandle = dlopen(binpath, RTLD_NOW)))
	{
		if((pAddr = dlsym(pHandle, "te")))
			te = *(ITempEntsSystem **)pAddr;
		else
			g_SMAPI->ConPrintf("** ERROR ** Tempents extension could not load correctly!\n");
		dlclose(pHandle);
	}
#endif
	GET_V_IFACE_CURRENT(engineFactory, enginetrace, IEngineTrace, INTERFACEVERSION_ENGINETRACE_SERVER);
	if(!enginetrace)
		ismm->ConPrintf("** ERROR ** Tempents extension could not load properly!\n");
	GET_V_IFACE_CURRENT(engineFactory, enginesound, IEngineSound, IENGINESOUND_SERVER_INTERFACE_VERSION);
	if(!enginesound)
		ismm->ConPrintf("** ERROR ** Tempents extension could not load properly!\n");
	GET_V_IFACE_ANY(serverFactory, gameents, IServerGameEnts, INTERFACEVERSION_SERVERGAMEENTS);
	if(!gameents)
		ismm->ConPrintf("** ERROR ** Tempents extension could not load properly!\n");
	return true;
}

enum
{
	TE_MASK_ALL=0,
	TE_MASK_SOLID,
	TE_MASK_PLAYERSOLID,
	TE_MASK_NPCSOLID,
	TE_MASK_WATER,
	TE_MASK_OPAQUE,
	TE_MASK_OPAQUE_AND_NPCS,
	TE_MASK_VISIBLE,
	TE_MASK_VISIBLE_AND_NPCS,
	TE_MASK_SHOT,
	TE_MASK_SHOT_HULL,
	TE_MASK_SHOT_PORTAL,
	TE_MASK_SOLID_BRUSHONLY,
	TE_MASK_PLAYERSOLID_BRUSHONLY,
	TE_MASK_NPCSOLID_BRUSHONLY,
	TE_MASK_NPCWORLDSTATIC,
	TE_MASK_SPLITAREAPORTAL,
	TE_MASK_CURRENT,
	TE_MASK_DEADSOLID
};

int EnumToMask(int enumval)
{
	switch(enumval)
	{
	case TE_MASK_ALL:
		return MASK_ALL;
		break;
	case TE_MASK_SOLID:
		return MASK_SOLID;
		break;
	case TE_MASK_PLAYERSOLID:
		return MASK_PLAYERSOLID;
		break;
	case TE_MASK_NPCSOLID:
		return MASK_NPCSOLID;
		break;
	case TE_MASK_WATER:
		return MASK_WATER;
		break;
	case TE_MASK_OPAQUE:
		return MASK_OPAQUE;
		break;
	case TE_MASK_OPAQUE_AND_NPCS:
		return MASK_OPAQUE_AND_NPCS;
		break;
	case TE_MASK_VISIBLE:
		return MASK_VISIBLE;
		break;
	case TE_MASK_VISIBLE_AND_NPCS:
		return MASK_VISIBLE_AND_NPCS;
		break;
	case TE_MASK_SHOT:
		return MASK_SHOT;
		break;
	case TE_MASK_SHOT_HULL:
		return MASK_SHOT_HULL;
		break;
	case TE_MASK_SOLID_BRUSHONLY:
		return MASK_SOLID_BRUSHONLY;
		break;
	case TE_MASK_PLAYERSOLID_BRUSHONLY:
		return MASK_PLAYERSOLID_BRUSHONLY;
		break;
	case TE_MASK_NPCSOLID_BRUSHONLY:
		return MASK_NPCSOLID_BRUSHONLY;
		break;
	case TE_MASK_NPCWORLDSTATIC:
		return MASK_NPCWORLDSTATIC;
		break;
	case TE_MASK_SPLITAREAPORTAL:
		return MASK_SPLITAREAPORTAL;
		break;
	case TE_MASK_CURRENT:
		return MASK_CURRENT;
		break;
	case TE_MASK_DEADSOLID:
		return MASK_DEADSOLID;
		break;
	default:
		return MASK_ALL;
		break;
	}
}

enum
{
	TE_FILTER_WORLD_ONLY=0,
	TE_FILTER_WORLD_AND_PROPS_ONLY,
	TE_FILTER_HIT_ALL
};

static cell_t TracelineNative(IPluginContext *pCtx, const cell_t *params)
{
	int numargs=params[0];
	if(numargs!=7)
		return -1;
	cell_t *startcell;
	pCtx->LocalToPhysAddr(params[1],&startcell);
	Vector start;
	start.x=sp_ctof(startcell[0]);
	start.y=sp_ctof(startcell[1]);
	start.z=sp_ctof(startcell[2]);
	cell_t *eyeangcell;
	pCtx->LocalToPhysAddr(params[2],&eyeangcell);
	QAngle eyes;
	eyes.x=sp_ctof(eyeangcell[0]);
	eyes.y=sp_ctof(eyeangcell[1]);
	eyes.z=sp_ctof(eyeangcell[2]);
	Vector eyevec;
	AngleVectors(eyes, &eyevec);
	eyevec.NormalizeInPlace();
	start += eyevec*64;
	Vector end = start + eyevec*10000000;
	Ray_t ray;
	ray.Init( start, end );
	trace_t tr;
	if(params[3]==TE_FILTER_WORLD_ONLY)
	{
		CTraceFilterWorldOnly traceFilter1;
		enginetrace->TraceRay(ray, EnumToMask(params[4]), &traceFilter1, &tr );
	}
	else if(params[3]==TE_FILTER_WORLD_AND_PROPS_ONLY)
	{
		CTraceFilterWorldAndPropsOnly traceFilter2;
		enginetrace->TraceRay(ray, EnumToMask(params[4]), &traceFilter2, &tr );
	}
	else
	{
		CTraceFilterHitAll traceFilter3;
		enginetrace->TraceRay(ray, EnumToMask(params[4]), &traceFilter3, &tr );
	}
	cell_t *vectoraddr;
	pCtx->LocalToPhysAddr(params[5],&vectoraddr);
	vectoraddr[0]=sp_ftoc(0.0);
	vectoraddr[1]=sp_ftoc(0.0);
	vectoraddr[2]=sp_ftoc(0.0);
	if(tr.fraction!=1)
	{
		vectoraddr[0]=sp_ftoc(tr.endpos.x);
		vectoraddr[1]=sp_ftoc(tr.endpos.y);
		vectoraddr[2]=sp_ftoc(tr.endpos.z);
	}
	if(sp_ctof(params[6])!=-1)
	{
		cell_t *frac;
		pCtx->LocalToPhysAddr(params[6],&frac);
		*frac=sp_ftoc(tr.fraction);
	}
	if(params[7]!=-1)
		if(tr.m_pEnt)
		{
			cell_t *ent;
			pCtx->LocalToPhysAddr(params[7],&ent);
			*ent=engine->IndexOfEdict(gameents->BaseEntityToEdict(tr.m_pEnt));
		}
	return 0;
}

static cell_t EmitSoundNative(IPluginContext *pCtx, const cell_t *params)
{
	MRecipientFilter filter;
	if(params[0]!=4)
		return -1;
	filter.AddRecipient(params[1]);
	char *eSound;
	pCtx->LocalToString(params[3],&eSound);
	if(!enginesound->IsSoundPrecached(eSound))
		enginesound->PrecacheSound(eSound, true);
	enginesound->EmitSound((IRecipientFilter&)filter,params[2],CHAN_AUTO,eSound,sp_ctof(params[4]),ATTN_NORM);
	return 1;
}

static cell_t ArmorRicochetNative(IPluginContext *pCtx, const cell_t *params)
{
	int numargs=params[0];
	if(numargs!=4)
		return -1;
	int index=params[1];
	float delay=sp_ctof(params[2]);
	cell_t *pos_cell;
	pCtx->LocalToPhysAddr(params[3],&pos_cell);
	Vector position;
	position.x=sp_ctof(pos_cell[0]);
	position.y=sp_ctof(pos_cell[1]);
	position.z=sp_ctof(pos_cell[2]);
	cell_t *dir_cell;
	pCtx->LocalToPhysAddr(params[4],&dir_cell);
	Vector direction;
	direction.x=sp_ctof(dir_cell[0]);
	direction.y=sp_ctof(dir_cell[1]);
	direction.z=sp_ctof(dir_cell[2]);
	MRecipientFilter filter;
	filter.AddPlayer(index);
	te->ArmorRicochet(filter,delay,&position,&direction);
	return 1;
}

static cell_t BeamPointsNative(IPluginContext *pCtx, const cell_t *params)
{
	int numargs=params[0];
	if(numargs!=14)
		return -1;
	int modelindex=params[1];
	int index=params[2];
	float delay=sp_ctof(params[3]);
	cell_t *start_cell;
	pCtx->LocalToPhysAddr(params[4],&start_cell);
	Vector start;
	start.x=sp_ctof(start_cell[0]);
	start.y=sp_ctof(start_cell[1]);
	start.z=sp_ctof(start_cell[2]);
	cell_t *end_cell;
	pCtx->LocalToPhysAddr(params[5],&end_cell);
	Vector end;
	end.x=sp_ctof(end_cell[0]);
	end.y=sp_ctof(end_cell[1]);
	end.z=sp_ctof(end_cell[2]);
	float life=sp_ctof(params[6]);
	float width=sp_ctof(params[7]);
	float endwidth=sp_ctof(params[8]);
	cell_t *color_cell;
	pCtx->LocalToPhysAddr(params[9],&color_cell);
	int r=color_cell[0];
	int g=color_cell[1];
	int b=color_cell[2];
	int a=color_cell[3];
	int startframe=params[10];
	int framerate=params[11];
	int fadelength=params[12];
	float aplitude=sp_ctof(params[13]);
	int speed=params[14];
	MRecipientFilter filter;
	filter.AddPlayer(index);
	te->BeamPoints(filter,delay,&start,&end,modelindex,0,startframe,framerate,life,width,endwidth,fadelength,aplitude,r,g,b,a,speed);
	return 1;
}

static cell_t BeamFollowNative(IPluginContext *pCtx, const cell_t *params)
{
	int numargs=params[0];
	if(numargs!=9)
		return -1;
	int modelindex=params[1];
	int index=params[2];
	int entindex=params[3];
	float delay=sp_ctof(params[4]);
	float life=sp_ctof(params[5]);
	float width=sp_ctof(params[6]);
	float endwidth=sp_ctof(params[7]);
	cell_t *color_cell;
	pCtx->LocalToPhysAddr(params[8],&color_cell);
	int r=color_cell[0];
	int g=color_cell[1];
	int b=color_cell[2];
	int a=color_cell[3];
	int fadelength=params[9];
	MRecipientFilter filter;
	filter.AddPlayer(index);
	te->BeamFollow(filter,delay,entindex,modelindex,0,life,width,endwidth,fadelength,r,g,b,a);
	return 1;
}

static cell_t BeamRingPointNative(IPluginContext *pCtx, const cell_t *params)
{
	int numargs=params[0];
	if(numargs!=14)
		return -1;
	int modelindex=params[1];
	int index=params[2];
	float delay=sp_ctof(params[3]);
	cell_t *pos_cell;
	pCtx->LocalToPhysAddr(params[4],&pos_cell);
	Vector position;
	position.x=sp_ctof(pos_cell[0]);
	position.y=sp_ctof(pos_cell[1]);
	position.z=sp_ctof(pos_cell[2]);
	float startradius=sp_ctof(params[5]);
	float endradius=sp_ctof(params[6]);
	float life=sp_ctof(params[7]);
	float width=sp_ctof(params[8]);
	cell_t *color_cell;
	pCtx->LocalToPhysAddr(params[9],&color_cell);
	int r=color_cell[0];
	int g=color_cell[1];
	int b=color_cell[2];
	int a=color_cell[3];
	int startframe=params[10];
	int framerate=params[11];
	float amplitude=sp_ctof(params[12]);
	int spread=params[13];
	int speed=params[14];
	MRecipientFilter filter;
	filter.AddPlayer(index);
	te->BeamRingPoint(filter,delay,position,startradius,endradius,modelindex,0,startframe,framerate,life,width,spread,amplitude,r,g,b,a,speed);
	return 1;
}

static cell_t SmokeNative(IPluginContext *pCtx, const cell_t *params)
{
	int numargs=params[0];
	if(numargs!=6)
		return -1;
	int modelindex=params[1];
	int index=params[2];
	float delay=sp_ctof(params[3]);
	cell_t *pos_cell;
	pCtx->LocalToPhysAddr(params[4],&pos_cell);
	Vector position;
	position.x=sp_ctof(pos_cell[0]);
	position.y=sp_ctof(pos_cell[1]);
	position.z=sp_ctof(pos_cell[2]);
	float scale=sp_ctof(params[5]);
	int framerate=params[6];
	MRecipientFilter filter;
	filter.AddPlayer(index);
	te->Smoke(filter,delay,&position,modelindex,scale,framerate);
	return 1;
}

static cell_t ExplosionNative(IPluginContext *pCtx, const cell_t *params)
{
	int numargs=params[0];
	if(numargs!=9)
		return -1;
	int modelindex=params[1];
	int index=params[2];
	float delay=sp_ctof(params[3]);
	cell_t *pos_cell;
	pCtx->LocalToPhysAddr(params[4],&pos_cell);
	Vector position;
	position.x=sp_ctof(pos_cell[0]);
	position.y=sp_ctof(pos_cell[1]);
	position.z=sp_ctof(pos_cell[2]);
	float scale=sp_ctof(params[5]);
	int framerate=params[6];
	int radius=params[7];
	int magnitude=params[8];
	cell_t *norm_cell;
	pCtx->LocalToPhysAddr(params[9],&norm_cell);
	Vector normal;
	normal.x=sp_ctof(norm_cell[0]);
	normal.y=sp_ctof(norm_cell[1]);
	normal.z=sp_ctof(norm_cell[2]);
	MRecipientFilter filter;
	filter.AddPlayer(index);
	te->Explosion(filter,delay,&position,modelindex,scale,framerate,0,radius,magnitude,&normal);
	return 1;
}

static cell_t MetalSparksNative(IPluginContext *pCtx, const cell_t *params)
{
	int numargs=params[0];
	if(numargs!=4)
		return -1;
	int index=params[1];
	float delay=sp_ctof(params[2]);
	cell_t *pos_cell;
	pCtx->LocalToPhysAddr(params[3],&pos_cell);
	Vector position;
	position.x=sp_ctof(pos_cell[0]);
	position.y=sp_ctof(pos_cell[1]);
	position.z=sp_ctof(pos_cell[2]);
	cell_t *dir_cell;
	pCtx->LocalToPhysAddr(params[4],&dir_cell);
	Vector direction;
	direction.x=sp_ctof(dir_cell[0]);
	direction.y=sp_ctof(dir_cell[1]);
	direction.z=sp_ctof(dir_cell[2]);
	MRecipientFilter filter;
	filter.AddPlayer(index);
	te->MetalSparks(filter,delay,&position,&direction);
	return 1;
}

const sp_nativeinfo_t tempents_natives[] = 
{
	{"TEArmorRicochet",				ArmorRicochetNative},
	{"TEBeamPoints",				BeamPointsNative},
	{"TEBeamFollow",				BeamFollowNative},
	{"TEBeamRingPoint",				BeamRingPointNative},
	{"TESmoke",						SmokeNative},
	{"TEMetalSparks",				MetalSparksNative},
	{"TETraceline",					TracelineNative},
	{"TEEmitSound",					EmitSoundNative},
	{"TEExplosion",					ExplosionNative},
	{NULL,							NULL},
};
