class Ext_SigScan
{
public:
#ifdef WIN32
	static void Init_Sigs();
#endif
};
#include "sigscan_class.h"

unsigned char* ustrncpy(unsigned char *dest, const unsigned char *src, int len)
{
	while(len--)
		dest[len] = src[len];
	return dest;
}

unsigned int istrcmp(char *str1, char *str2)
{
	unsigned int i, len1 = strlen(str1), len2 = strlen(str2);
	if(len1 != len2)
		return 0;
	for(i = 0;i < len1;i++)
	{
		if(tolower(str1[i]) != tolower(str2[i]))
			return 0;
	}
	return 1;
}

#ifdef WIN32
unsigned char* SigScan::base_addr;
size_t SigScan::base_len;
void SigScan::Init(char *name, unsigned char *sig, char *mask, size_t len)
{
	strncpy(sig_name, name, 63);
	sig_name[64] = '\0';
	is_set = 0;
	sig_len = len;
	sig_str = new unsigned char[sig_len];
	ustrncpy(sig_str, sig, sig_len);
	sig_mask = new char[sig_len+1];
	strncpy(sig_mask, mask, sig_len);
	sig_mask[sig_len+1] = 0;
	if(!base_addr)
	{
		Msg("Failed to find the server module, plugin will not function properly.\n");
		return ;
	}
	if((sig_addr = FindSignature()) == NULL)
	{
		Msg("Failed to find the signature \"%s\", plugin will not function properly.\n",this->sig_name);
		return ;
	}
	is_set = 1;
	g_SMAPI->LogMsg(mmplugin,"Found signature function: %s\n", this->sig_name);
	return ;
}

SigScan::~SigScan(void)
{
	delete[] sig_str;
	delete[] sig_mask;
}

bool SigScan::GetDllMemInfo(void)
{
	char binpath[1024];
	HANDLE hModuleSnap = INVALID_HANDLE_VALUE;
	MODULEENTRY32 modent;
	base_addr = 0;
	base_len = 0;
	engine->GetGameDir(binpath, 512);
	sprintf(binpath, "%s\\bin\\server.dll", binpath);
	hModuleSnap = CreateToolhelp32Snapshot(TH32CS_SNAPMODULE, GetCurrentProcessId());
	if(hModuleSnap == INVALID_HANDLE_VALUE)
	{
		Msg("CreateToolhelp32Snapshot failed.\n");
		return false;
	}
	modent.dwSize = sizeof(MODULEENTRY32);
	if(!Module32First(hModuleSnap, &modent))
	{
		Msg("Module32First failed.\n");
		CloseHandle(hModuleSnap);
		return false;
	}
	do
	{
		if(istrcmp(modent.szExePath, binpath))
		{
			base_addr = modent.modBaseAddr;
			base_len = modent.modBaseSize;
			CloseHandle(hModuleSnap);
			return true;
		}
	} while(Module32Next(hModuleSnap, &modent));
	CloseHandle(hModuleSnap);
	Msg("Failed to find server module in module list.\n");
	return false;
}

void* SigScan::FindSignature(void)
{
	unsigned char *pBasePtr = base_addr;
	unsigned char *pEndPtr = base_addr+base_len;
	size_t i, height = 0;
	while(pBasePtr < pEndPtr)
	{
		for(i = 0;i < sig_len;i++)
		{
			if(i > height)
				height++;
			if((sig_mask[i] != '?') && (sig_str[i] != pBasePtr[i]))
				break;
		}
		if(i == sig_len)
			return (void*)pBasePtr;
		pBasePtr++;
	}

	Msg("Sig Failed at Height: %d\n", height);

	return NULL;
}

void Ext_SigScan::Init_Sigs(void) {
	SigScan::GetDllMemInfo();
	return ;
}
#endif /* if WIN32 */
#ifndef WIN32 /* !WIN32 */
SigScan::~SigScan(void)
{
}
void find_sym_addr(void *dl_handle, char *name, char *symbol, void **value) {
	void *addr;

	if(dl_handle == NULL)
		return;

	addr = dlsym(dl_handle, symbol);
	if(addr == NULL) {
		Msg("dlsym() failed for function \"%s\" with error: \"%s\", plugin will not function properly.",
			name, dlerror());
		return;
	}

	Msg("Found linux symbol function: %s, address: %X\n", name, addr);
	*value=addr;
}
#endif /* endif !WIN32 */
